% psi = pswf(x, T, W, m, eps)
% 
% Generates m Prolate Spheroidal Wave Functions in time-bandwidth T and W 
% evaluated at x.
% 
% Input:
%   x   - A 1 by k size vector of evaluation points. Values must be between 
%         -1 and 1.
%   T   - Time limit of functions to be generated.
%   W   - Bandwidth of functions to be generated.
%   m   - (optional) Number of basis functions to return.
%   eps - (optional) Min precision of computed pswfs.
% 
% Output:
%   psi - Prolate Spheroidal wave functions. A m by k matrix where every
%         row is a pswf.
% 
% Reference:
%   H. Xiao, ``Prolate Spheroidal Wave Functions, Quadrature,
%   Interpolation, and Asymptotic Formulae.'' PhD thesis, Yale University,
%   2001, pp. 19-21.
% 
function psi = pswf(x, T, W , m, eps)
    if nargin<5
        eps = 1e-4;
    end
    c = pi*T*W;
    if nargin<4
        n = ceil(2*(exp(1)*c+1)+log2(1/eps));
        m = n;
    else
        n = max(ceil(2*(exp(1)*c+1)+log2(1/eps)), m);
    end
    
    t = (0:n).';
    am1 = (t.*(t-1))./((2*t-1).*sqrt((2*t-3).*(2*t+1)))*c^2;
    a0 = t.*(t+1)+(2*t.*(t+1)-1)./((2*t+3).*(2*t-1))*c^2;
    a1 = ((t+2).*(t+1))./((2*t+3).*sqrt((2*t+1).*(2*t+5)))*c^2;

    H = spdiags([a1 a0 am1], [-2 0 2], n+1, n+1);
    H = H/n^2;
    
    [U, e] = eig(full(H));
    [~, jj] = sort(diag(e));
    U = U(:, jj);
    
    P = legendrePoly(n-1,x);
    
    psi = U(1:n,1:m)'*P;
    for k=1:size(psi,1)
        psi(k,:) = psi(k,:)/norm(psi(k,:));
    end
end


% P = legendrePoly(n, X)
% 
% Compute the first n normalized Legendre polynomials
% 
% Input:
%   n - Legendre polynomial to compute up to. (n>=0)
%   X - Evaluation points for Legendre polynomials. Must be a 1xm vector 
%       with values between -1 and 1.
% Output:
%   P - First n normalized Legendre polynomials. Is a n by m matrix where
%       every right is a Legendre polynomial evaluated at X.
% 
% Reference:
%   H. Xiao, ``Prolate Spheroidal Wave Functions, Quadrature,
%   Interpolation, and Asymptotic Formulae.'' PhD thesis, Yale University,
%   2001, pp. 5-6.
% 
function P = legendrePoly(n, X)
    if n<0
        error('n must be non-negative');
    end
    P = zeros(n+1, length(X));
    
    P(1,:) = ones(1,length(X));
    if n==0
        return;
    end
    
    P(2,:) = X;
    if n==1
        return;
    end
    
    for k=2:n
        P(k+1,:) = (2*k-1)*X.*P(k,:) - (k-1)*P(k-1,:);
        P(k+1,:) = P(k+1,:)/k;
    end
    
    % normalize them
    for k=0:n
        P(k+1,:) = P(k+1,:)*sqrt(k+0.5);
    end
end